from _typeshed import Incomplete

class InvalidColor(Exception):
    message: Incomplete
    def __init__(self, message: str) -> None: ...

class InvalidHexColor(Exception):
    message: Incomplete
    def __init__(self, message: str) -> None: ...

class InvalidStyle(Exception):
    message: Incomplete
    def __init__(self, message: str) -> None: ...

class InvalidControl(Exception):
    message: Incomplete
    def __init__(self, message: str) -> None: ...
