from .hexadecimal import Hex as Hex
from .library import Library as Library
from .utilities import Utilities as Utilities
from _typeshed import Incomplete
from typing import Any

TTY_AWARE: bool
IS_TTY: Incomplete

class Colored:
    def __init__(self, name: Any) -> None: ...
    def attribute(self, line_color: str = '') -> str: ...
    def foreground(self) -> str: ...
    def background(self) -> str: ...
    @staticmethod
    def enable_windows_terminal_mode() -> Any: ...
    @staticmethod
    def enabled() -> bool: ...

def style(name: int | str, color: str | int = '') -> str: ...
def fore(name: int | str) -> str: ...
def back(name: int | str) -> str: ...
def fore_rgb(r: int | str, g: int | str, b: int | str) -> str: ...
def back_rgb(r: int | str, g: int | str, b: int | str) -> str: ...
def attr(name: int | str) -> str: ...
def fg(name: int | str) -> str: ...
def bg(name: int | str) -> str: ...
def stylize(text: str, formatting: int | str, reset: bool = True) -> str: ...
def stylize_interactive(text: str, formatting: str, reset: bool = True) -> str: ...
def set_tty_aware(awareness: bool = True) -> None: ...
