from .attributes import Style as Style
from .background import Back as Back
from .colored import Colored as Colored, attr as attr, back as back, back_rgb as back_rgb, bg as bg, fg as fg, fore as fore, fore_rgb as fore_rgb, set_tty_aware as set_tty_aware, style as style, stylize as stylize, stylize_interactive as stylize_interactive
from .controls import Controls as Controls
from .cprint import cprint as cprint
from .foreground import Fore as Fore

__version__: str
