from .exceptions import InvalidColor as InvalidColor, InvalidControl as InvalidControl, InvalidHexColor as InvalidHexColor, InvalidStyle as InvalidStyle
from .library import Library as Library

class Utilities:
    RGB_MAXIMUM_COLOR: int
    colorterm: str
    def __init__(self) -> None: ...
    def is_color_exist(self, name: str) -> bool: ...
    def is_style_exist(self, name: str) -> bool: ...
    def is_control_exist(self, name: str) -> bool: ...
    def convert_percentages(self, percent: str | int) -> int | str: ...
    def set_colorterm(self, colorterm: str = ''): ...
    def is_percentage(self, numbers: tuple) -> list: ...
