from .library import Library as Library

class Hex:
    def find(self, color: str | int) -> str: ...
    @staticmethod
    def cube(x: int) -> int: ...
    def fit(self, hex_val: str, ref: int) -> int: ...
